/****************************************************************************
** $Id: qsdate_object.cpp  20030513   edited May 12 18:57 $
**
** Copyright (C) 2001-2003 Trolltech AS.  All rights reserved.
**
** This file is part of the Qt Script for Applications framework (QSA).
**
** This file may be distributed and/or modified under the terms of the
** GNU General Public License version 2 as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL included in the
** packaging of this file.
**
** Licensees holding a valid QSA Beta Evaluation Version license may use
** this file in accordance with the QSA Beta Evaluation Version License
** Agreement provided with the Software.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** See http://www.trolltech.com/pricing.html or email sales@trolltech.com for
**   information about QSA Commercial License Agreements.
** See http://www.trolltech.com/gpl/ for GPL licensing information.
**
** Contact info@trolltech.com if any conditions of this licensing are
** not clear to you.
**
*****************************************************************************/

#include "qsdate_object.h"

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include "qsoperations.h"
#include <qdatetime.h>

#ifdef Q_OS_UNIX
#  include <unistd.h>
#  include <sys/time.h>
#endif

#include <math.h>
#include <string.h>
#include <stdio.h>
#include <locale.h>
#include <time.h>
#ifdef HAVE_SYS_TIME_H
#  include <unistd.h>
#  include <sys/time.h>
#endif
#ifdef HAVE_SYS_TIMEB_H
#  include <sys/timeb.h>
#endif
#ifdef HAVE_SYS_PARAM_H
#  include <sys/param.h>
#endif

enum { ToString, ToDateString, ToTimeString, ToLocaleString,
       ToLocaleDateString, ToLocaleTimeString, ValueOf, GetTime,
       GetFullYear, GetMonth, GetDate, GetDay, GetHours, GetMinutes,
       GetSeconds, GetMilliSeconds, GetTimezoneOffset, SetTime,
       SetMilliSeconds, SetSeconds, SetMinutes, SetHours, SetDate,
       SetMonth, SetFullYear, ToUTCString,
       // non-normative properties (Appendix B)
       GetYear, SetYear, ToGMTString,
       UTCOffset = 100 };

// helper functions
QSObject parseDate( QSEnv *env, const QSString &s);
QSObject timeClip(const QSObject &t);

QSObject parseDate( QSEnv *env, const QSString &s)
{
    QString u = s.value();
    int firstSlash = u.find('/');
    if ( firstSlash == -1 )
	{
	    /* TODO parse dates like "December 25, 1995 23:15:00"*/
	    qWarning( "QtScript: parsing this date format isn't implemented: %s", u.ascii() );
	    return env->createNumber(0);
	}
    else
	{
	    // Found 12/31/2099 on some website -> obviously MM/DD/YYYY
	    int month = QSString::toULong( u.mid( 0,firstSlash ) );
	    int secondSlash = u.find('/',firstSlash+1);
	    if ( secondSlash == -1 )
		{
		    qWarning( "QtScript: parsing this date format isn't implemented: %s", u.ascii() );
		    return env->createNumber(0);
		}
	    int day = QSString::toULong( u.mid( firstSlash+1,
						secondSlash-firstSlash-1) );
	    int year = QSString::toULong( u.mid( secondSlash+1 ) );
	    struct tm t;
	    memset( &t, 0, sizeof(t) );
	    year = (year > 2037) ? 2037 : year; // mktime is limited to 2037 !!!
	    t.tm_year = (year >= 0 && year <= 99) ? year : year - 1900;
	    t.tm_mon = month-1; // mktime wants 0-11 for some reason
	    t.tm_mday = day;
	    time_t seconds = mktime(&t);
	    if ( seconds == -1 )
		{
		    qWarning( "QtScript: mktime returned -1: %s", u.ascii() );
		    return env->createUndefined();
		}
	    else
		return env->createNumber( seconds * 1000.0 );
	}
}

QSObject timeClip(const QSObject &t)
{
    /* TODO */
    return t;
}

QSDateClass::QSDateClass( QSClass *b )
    : QSClass( b, AttributeFinal )
{
    addMember( "parse", QSMember( &parse, AttributeStatic ) );
    addMember( "utc", QSMember( &utc, AttributeStatic ) );

    addMember( "toString", QSMember( &QSDateClass::toString ) );
    addMember( "toDateString", QSMember( &toDateString ) );
    addMember( "toTimeString", QSMember( &toTimeString ) );
    addMember( "toLocaleString", QSMember( &toLocaleString ) );
    addMember( "toLocaleDateString", QSMember( &toLocaleDateString ) );
    addMember( "valueOf", QSMember( &valueOf ) );

    addMember( "getTime", QSMember( &getTime ) );
    addMember( "getFullYear", QSMember( &getFullYear ) );
    addMember( "getUTCFullYear", QSMember( &getFullYear ) );
    addMember( "getMonth", QSMember( &getMonth ) );
    addMember( "getUTCMonth", QSMember( &getMonth ) );
    addMember( "getDate", QSMember( &getDate ) );
    addMember( "getUTCDate", QSMember( &getDate ) );
    addMember( "getMinutes", QSMember( &getMinutes ) );
    addMember( "getUTCMinutes", QSMember( &getMinutes ) );
    addMember( "getSeconds", QSMember( &getSeconds ) );
    addMember( "getUTCSeconds", QSMember( &getSeconds ) );
    addMember( "getMilliSeconds", QSMember( &getMilliSeconds ) );
    addMember( "getUTCMilliSeconds", QSMember( &getMilliSeconds ) );
    addMember( "getTimezoneOffset", QSMember( &getTimezoneOffset ) );

    addMember( "setTime", QSMember( &setTime ) );
    addMember( "setMilliseconds", QSMember( &setMilliseconds ) );
    addMember( "setUTCMilliseconds", QSMember( &setMilliseconds ) );
    addMember( "setHours", QSMember( &setHours ) );
    addMember( "setSeconds", QSMember( &setSeconds ) );
    addMember( "setUTCSeconds", QSMember( &setSeconds ) );
    addMember( "setMinutes", QSMember( &setMinutes ) );
    addMember( "setUTCMinutes", QSMember( &setMinutes ) );
    addMember( "setDate", QSMember( &setDate ) );
    addMember( "setUTCDate", QSMember( &setDate ) );
    addMember( "setMonth", QSMember( &setMonth ) );
    addMember( "setUTCMonth", QSMember( &setMonth ) );
    addMember( "setFullYear", QSMember( &setFullYear ) );
    addMember( "setUTCFullYear", QSMember( &setFullYear ) );
    addMember( "setYear", QSMember( &setYear ) );

    addMember( "getYear", QSMember( &getYear ) );
    addMember( "toGMTString", QSMember( &toGMTString ) );
}

QSObject QSDateClass::construct( const QSList &args ) const
{
    double value;

    int numArgs = args.size();

    if (numArgs == 0) { // new Date() ECMA 15.9.3.3
#ifdef HAVE_SYS_TIMEB_H
#  if defined(__BORLANDC__)
	struct timeb timebuffer;
	ftime(&timebuffer);
#  else
	struct _timeb timebuffer;
	_ftime(&timebuffer);
#  endif
	double utc = floor((double)timebuffer.time * 1000.0 + (double)timebuffer.millitm);
#else
	struct timeval tv;
	gettimeofday(&tv, 0L);
	double utc = floor((double)tv.tv_sec * 1000.0 + (double)tv.tv_usec / 1000.0);
#endif // HAVE_SYS_TIMEB_H
	value = utc;
    } else if (numArgs == 1) {
	QSObject p = args[0].toPrimitive();
	if ( p.isString() )
	    value = parseDate( env(), QSString( env(), p.toString() ) ).toNumber();
	else
	    value = p.toNumber();
    } else {
	struct tm t;
	memset(&t, 0, sizeof(t));
	QSObject y = createNumber( args[0].toNumber() );
	/* TODO: check for NaN */
	int year = y.toInt32();
	t.tm_year = (year >= 0 && year <= 99) ? year : year - 1900;
	t.tm_mon = args[1].toInt32();
	t.tm_mday = (numArgs >= 3) ? args[2].toInt32() : 1;
	t.tm_hour = (numArgs >= 4) ? args[3].toInt32() : 0;
	t.tm_min = (numArgs >= 5) ? args[4].toInt32() : 0;
	t.tm_sec = (numArgs >= 6) ? args[5].toInt32() : 0;
	t.tm_isdst = -1;
	int ms = (numArgs >= 7) ? args[6].toInt32() : 0;
	value = mktime(&t) * 1000.0 + ms;
    }

    QSObject date( this );
    date.setVal( value );
    return date;
}

QSObject QSDateClass::cast( const QSList & ) const
{
    time_t t = time( 0L );
    QString s( ctime( &t ) );

    // return formatted string minus trailing \n
    return createString( s.mid(0, s.length() - 1 ) );
}

bool QSDateClass::toBoolean( const QSObject *obj ) const
{
    return bool(obj->dVal());
}

double QSDateClass::toNumber( const QSObject *obj ) const
{
    return obj->dVal();
}

QString QSDateClass::toString( const QSObject *o ) const
{
    double milli = o->dVal();
    time_t tv = (time_t) floor(milli / 1000.0);
    // struct tm *t = localtime(&tv);
    QString s = ctime( &tv );
    return s.mid( 0, s.length() - 1 );
}

QVariant QSDateClass::toVariant( const QSObject *obj, QVariant::Type t ) const
{
    // ### simplify
    QSList args;
    if ( t == QVariant::Time ) {
	QTime time( obj->get( "getHours").execute( args ).toInteger(),
		    obj->get( "getMinutes").execute( args ).toInteger(),
		    obj->get( "getSeconds").execute( args ).toInteger(),
		    obj->get( "getMilliseconds").execute( args ).toInteger() );
	return time;
    } else if ( t == QVariant::DateTime ) {
	QTime time( obj->get( "getHours").execute( args ).toInteger(),
		    obj->get( "getMinutes").execute( args ).toInteger(),
		    obj->get( "getSeconds").execute( args ).toInteger(),
		    obj->get( "getMilliseconds").execute( args ).toInteger() );
	QDate date( obj->get( "getFullYear").execute( args ).toInteger(),
		    obj->get( "getMonth").execute( args ).toInteger() + 1,
		    obj->get( "getDate").execute( args ).toInteger() );
	return QDateTime( date, time );
    } else {
	QDate date( obj->get( "getFullYear").execute( args ).toInteger(),
		    obj->get( "getMonth").execute( args ).toInteger() + 1,
		    obj->get( "getDate").execute( args ).toInteger() );
	return date;
    }
}

// Date.parse()
QSObject QSDateClass::parse( QSEnv *env )
{
    if ( env->arg( 0 ).isA( env->stringClass() ) )
	return parseDate( env, QSString( env, env->arg( 0 ).toString() ) );
    else
	return env->createUndefined();
}

// Date.UTC()
QSObject QSDateClass::utc( QSEnv *env )
{
    struct tm t;
    memset(&t, 0, sizeof(t));
    const QSList *args = env->arguments();
    int n = args->size();
    /* TODO: check for NaN */
    int year = env->arg( 0 ).toInt32();
    t.tm_year = (year >= 0 && year <= 99) ? year : year - 1900;
    t.tm_mon = env->arg( 1 ).toInt32();
    t.tm_mday = (n >= 3) ? env->arg( 2 ).toInt32() : 1;
    t.tm_hour = (n >= 4) ? env->arg( 3 ).toInt32() : 0;
    t.tm_min = (n >= 5) ? env->arg( 4 ).toInt32() : 0;
    t.tm_sec = (n >= 6) ? env->arg( 5 ).toInt32() : 0;
    int ms = (n >= 7) ? env->arg( 6 ).toInt32() : 0;

    return env->createNumber( mktime(&t) * 1000.0 + ms );
}

QSObject QSDateClass::toString( QSEnv *env )
{
    return QSDateClass::invokeS( env, ToString );
}

QSObject QSDateClass::toDateString( QSEnv *env )
{
    return QSDateClass::invokeS( env, ToDateString );
}

QSObject QSDateClass::toTimeString( QSEnv *env )
{
    return QSDateClass::invokeS( env, ToTimeString );
}

QSObject QSDateClass::toGMTString( QSEnv *env )
{
    return QSDateClass::invokeS( env, ToGMTString );
}

QSObject QSDateClass::toLocaleString( QSEnv *env )
{
    return QSDateClass::invokeS( env, ToLocaleString );
}

QSObject QSDateClass::toLocaleDateString( QSEnv *env )
{
    return QSDateClass::invokeS( env, ToLocaleDateString );
}

QSObject QSDateClass::toLocaleTimeString( QSEnv *env )
{
    return QSDateClass::invokeS( env, ToLocaleTimeString );
}

QSObject QSDateClass::valueOf( QSEnv *env )
{
    return QSDateClass::invokeS( env, ValueOf );
}

QSObject QSDateClass::getTime( QSEnv *env )
{
    return QSDateClass::invokeS( env, GetTime );
}

QSObject QSDateClass::getYear( QSEnv *env )
{
    return QSDateClass::invokeS( env, GetYear );
}

QSObject QSDateClass::getFullYear( QSEnv *env )
{
    return QSDateClass::invokeS( env, GetFullYear );
}

QSObject QSDateClass::getMonth( QSEnv *env )
{
    return QSDateClass::invokeS( env, GetMonth );
}

QSObject QSDateClass::getDate( QSEnv *env )
{
    return QSDateClass::invokeS( env, GetDate );
}

QSObject QSDateClass::getMinutes( QSEnv *env )
{
    return QSDateClass::invokeS( env, GetMinutes );
}

QSObject QSDateClass::getSeconds( QSEnv *env )
{
    return QSDateClass::invokeS( env, GetSeconds );
}

QSObject QSDateClass::getMilliSeconds( QSEnv *env )
{
    return QSDateClass::invokeS( env, GetMilliSeconds );
}

QSObject QSDateClass::getTimezoneOffset( QSEnv *env )
{
    return QSDateClass::invokeS( env, GetTimezoneOffset );
}

QSObject QSDateClass::setTime( QSEnv *env )
{
    return QSDateClass::invokeS( env, SetTime );
}

QSObject QSDateClass::setMilliseconds( QSEnv *env )
{
    return QSDateClass::invokeS( env, SetMilliSeconds );
}

QSObject QSDateClass::setSeconds( QSEnv *env )
{
    return QSDateClass::invokeS( env, SetSeconds );
}

QSObject QSDateClass::setMinutes( QSEnv *env )
{
    return QSDateClass::invokeS( env, SetMinutes );
}

QSObject QSDateClass::setHours( QSEnv *env )
{
    return QSDateClass::invokeS( env, SetHours );
}

QSObject QSDateClass::setDate( QSEnv *env )
{
    return QSDateClass::invokeS( env, SetDate );
}

QSObject QSDateClass::setMonth( QSEnv *env )
{
    return QSDateClass::invokeS( env, SetMonth );
}

QSObject QSDateClass::setFullYear( QSEnv *env )
{
    return QSDateClass::invokeS( env, SetFullYear );
}

QSObject QSDateClass::setYear( QSEnv *env )
{
    return QSDateClass::invokeS( env, SetYear );
}

QSObject QSDateClass::invokeS( QSEnv *env, int id )
{
    bool utc = FALSE;
    if ( id >= UTCOffset ) {
	utc = TRUE;
	id -= UTCOffset;
    }

    QSObject result;
    QString s;
    const int bufsize=100;
    char timebuffer[bufsize];
    char *oldlocale = setlocale(LC_TIME,NULL);
    if (!oldlocale)
	oldlocale = setlocale(LC_ALL, NULL);
    QSObject thisObj = env->thisValue();
    double milli = thisObj.toNumber();
    time_t tv = (time_t) floor(milli / 1000.0);
    int ms = int(milli - tv * 1000.0);

    struct tm *t;
    if (utc)
	t = gmtime(&tv);
    else
	t = localtime(&tv);

    // trick gcc. We don't want the Y2K warnings.
    const char *xFormat = "%x";
    const char *cFormat = "%c";

    switch (id) {
    case ToString:
	s = ctime(&tv);
	result = env->createString(s.mid(0, s.length() - 1));
	break;
    case ToDateString:
    case ToTimeString:
    case ToGMTString:
	setlocale( LC_TIME, "C" );
	if ( id == ToDateString ) {
	    strftime(timebuffer, bufsize, xFormat, t);
	} else if ( id == ToTimeString ) {
	    strftime(timebuffer, bufsize, "%X",t);
	} else {
	    t = gmtime(&tv);
	    strftime(timebuffer, bufsize, "%a, %d %b %Y %H:%M:%S %z", t);
	}
	setlocale(LC_TIME,oldlocale);
	result = env->createString( timebuffer );
	break;
    case ToLocaleString:
	strftime(timebuffer, bufsize, cFormat, t);
	result = env->createString( timebuffer );
	break;
    case ToLocaleDateString:
	strftime(timebuffer, bufsize, xFormat, t);
	result = env->createString( timebuffer );
	break;
    case ToLocaleTimeString:
	strftime(timebuffer, bufsize, "%X", t);
	result = env->createString( timebuffer );
	break;
    case ValueOf:
    case GetTime:
	result = env->createNumber( milli );
	break;
    case GetYear:
	result = env->createNumber( t->tm_year );
	break;
    case GetFullYear:
	result = env->createNumber( 1900 + t->tm_year );
	break;
    case GetMonth:
	result = env->createNumber( t->tm_mon );
	break;
    case GetDate:
	result = env->createNumber( t->tm_mday );
	break;
    case GetDay:
	result = env->createNumber( t->tm_wday );
	break;
    case GetHours:
	result = env->createNumber( t->tm_hour );
	break;
    case GetMinutes:
	result = env->createNumber( t->tm_min );
	break;
    case GetSeconds:
	result = env->createNumber( t->tm_sec );
	break;
    case GetMilliSeconds:
	result = env->createUndefined();
	break;
    case GetTimezoneOffset:
#if defined BSD || defined(__APPLE__)
	result = env->createNumber(-( t->tm_gmtoff / 60 ) + ( t->tm_isdst ? 60 : 0 ));
#else
#  if defined(__BORLANDC__)
// ### verify daylight offset
	result = env->createNumber(_timezone / 60 - (_daylight ? 60 : 0));
#  else
	result = env->createNumber(( timezone / 60 - ( daylight ? 60 : 0 )));
#  endif
#endif
	break;
    case SetTime:
	milli = env->arg( 0 ).round();
	result = env->createNumber( milli );
	//###   thisObj.setInternalValue(result);
	break;
    case SetMilliSeconds:
	ms = env->arg( 0 ).toInt32();
	break;
    case SetSeconds:
	t->tm_sec = env->arg( 0 ).toInt32();
	break;
    case SetMinutes:
	t->tm_min = env->arg( 0 ).toInt32();
	break;
    case SetHours:
	t->tm_hour = env->arg( 0 ).toInt32();
	break;
    case SetDate:
	t->tm_mday = env->arg( 0 ).toInt32();
	break;
    case SetMonth:
	t->tm_mon = env->arg( 0 ).toInt32();
	break;
    case SetFullYear:
	t->tm_year = env->arg( 0 ).toInt32() - 1900;
	break;
    case SetYear:
	t->tm_year = env->arg( 0 ).toInt32() >= 1900 ? env->arg( 0 ).toInt32() - 1900 : env->arg( 0 ).toInt32();
	break;
    }

    if (id == SetYear || id == SetMilliSeconds || id == SetSeconds ||
	id == SetMinutes || id == SetHours || id == SetDate ||
	id == SetMonth || id == SetFullYear ) {
	result = env->createNumber( mktime(t) * 1000.0 + ms );
	// ### thisObj.setInternalValue(result);
    }

    return result;
}
